package iface

import (
	"math"
	"sort"
	"time"
)

type RequestDurationAggr struct {
	lastMinuteRequestTimes []float64
	Min                    float64   `json:"min"`
	Max                    float64   `json:"max"`
	Mean                   float64   `json:"mean"`
	Stdev                  float64   `json:"stdev"`
	P90                    float64   `json:"p90"`
	P95                    float64   `json:"p95"`
	P99                    float64   `json:"p99"`
	Timestamp              time.Time `json:"timestamp"`
}

// NewRequestDurationAggr returns a new initialized RequestDurationAggr
// object.
func NewRequestDurationAggr() *RequestDurationAggr {
	rt := &RequestDurationAggr{}
	rt.lastMinuteRequestTimes = make([]float64, 0)
	rt.Timestamp = time.Now()
	return rt
}

// StartTimer will call a forever loop in a goroutine to calculate
// metrics for measurements every d ticks.
func (rt *RequestDurationAggr) StartTimer(d time.Duration) {
	timer := time.Tick(d)
	go func() {
		for {
			<-timer
			rt.calculate()
		}
	}()
}

// GetStats to fulfill Metrics.Metric interface, it returns the data
// that will be served as JSON.
func (rt *RequestDurationAggr) GetStats() interface{} {
	return rt
}

// Name to fulfill Metrics.Metric interface, it will return the name
// of the JSON object that will be served.
func (rt *RequestDurationAggr) Name() string {
	return "RequestTime"
}

// InRoot to fulfill Metrics.Metric interface, it will return where to
// put the JSON object into the monitoring endpoint.
func (rt *RequestDurationAggr) InRoot() bool {
	return false
}

func (rt *RequestDurationAggr) Add(n float64) {
	rt.lastMinuteRequestTimes = append(rt.lastMinuteRequestTimes, n)
}

func (rt *RequestDurationAggr) calculate() {
	sortedSlice := rt.lastMinuteRequestTimes[:]
	rt.lastMinuteRequestTimes = make([]float64, 0)
	l := len(sortedSlice)
	if l <= 1 {
		return
	}
	sort.Float64s(sortedSlice)

	rt.Timestamp = time.Now()
	rt.Min = sortedSlice[0]
	rt.Max = sortedSlice[l - 1]
	rt.Mean = mean(sortedSlice, l)
	rt.Stdev = correctedStdev(sortedSlice, rt.Mean, l)
	rt.P90 = p90(sortedSlice, l)
	rt.P95 = p95(sortedSlice, l)
	rt.P99 = p99(sortedSlice, l)
}

func mean(orderedObservations []float64, l int) float64 {
	res := 0.0
	for i := 0; i < l; i++ {
		res += orderedObservations[i]
	}

	return res / float64(l)
}

func p90(orderedObservations []float64, l int) float64 {
	return percentile(orderedObservations, l, 0.9)
}

func p95(orderedObservations []float64, l int) float64 {
	return percentile(orderedObservations, l, 0.95)
}

func p99(orderedObservations []float64, l int) float64 {
	return percentile(orderedObservations, l, 0.99)
}

// percentile with argument p \in (0,1), l is the length of given orderedObservations
// It does a simple apporximation of an ordered list of observations.
// Formula: sortedSlice[0.95*length(sortedSlice)]
func percentile(orderedObservations []float64, l int, p float64) float64 {
	return orderedObservations[int(p * float64(l))]
}

func correctedStdev(observations []float64, mean float64, l int) float64 {
	var omega float64
	for i := 0; i < l; i++ {
		omega += math.Pow(observations[i]-mean, 2)
	}
	stdev := math.Sqrt(1 / (float64(l) - 1) * omega)
	return stdev
}
