// +build go1.8

package http

import (
	"time"
	"net/http"

	"amuz.es/gogs/infra/api_skeleton/subsys/http/middleware/logging"
	"amuz.es/gogs/infra/api_skeleton/subsys/http/middleware/recovery"
	"amuz.es/gogs/infra/api_skeleton/util"
	"github.com/gin-gonic/gin"
	errs "github.com/pkg/errors"

	"amuz.es/gogs/infra/api_skeleton/subsys/http/route"
	"github.com/mkevac/debugcharts"
	"log"
	"context"
	"sync"
)

var (
	logger = util.NewLogger("http")
)

func Start(bindAddress string, accessLogConfig *util.LogConfig,
	errorSignal chan error, ctx context.Context, waiter *sync.WaitGroup) {
	server, err := initialize(bindAddress, accessLogConfig)
	if err != nil {
		errorSignal <- err
		return
	}
	go bind(server, errorSignal, waiter)
	go gracefulCloser(server, ctx, waiter)
	logger.Info(bindAddress + " bound")

}
func initialize(bindAddress string, accessLogConfig *util.LogConfig) (server *http.Server, err error) {
	defer func() {
		if recov := recover(); recov != nil {
			err = recov.(error)
		}
	}()
	gin.SetMode(gin.ReleaseMode)
	engine := gin.New()

	initMiddleware(engine, accessLogConfig)
	initRoutes(engine)

	logWriter := logger.Writer()
	server = &http.Server{
		Addr:           bindAddress,
		Handler:        engine,
		ReadTimeout:    10 * time.Second,
		WriteTimeout:   10 * time.Second,
		IdleTimeout:    60 * time.Second,
		MaxHeaderBytes: 1 << 20,
		ErrorLog:       log.New(logWriter, "", 0),
	}
	return
}
func gracefulCloser(server *http.Server, ctx context.Context, waiter *sync.WaitGroup) {
	defer waiter.Done()
	<-ctx.Done()
	logger.Info("stopping HTTP")
	if err := server.Shutdown(context.Background()); err != nil {
		logger.Info("HTTP closed:", err)
	}
	logger.Info("HTTP shutdowned safely.")
}

func bind(server *http.Server, errorSignal chan error, waiter *sync.WaitGroup) {
	waiter.Add(1)
	if err := server.ListenAndServe(); err != nil {
		errorSignal <- errs.Wrap(err, "failed to listen or serve http")
	}
}

func initRoutes(engine *gin.Engine) {
	suburl := ""
	path := engine.Group(suburl)
	// handlers

	//debug handler
	debug := path.Group("/debug")
	debug.GET("/metric", route.Metric)
	debugcharts.GinDebugRouter(engine)
	debug.GET("/pprof/", route.ProfileIndexHandler())
	debug.GET("/pprof/heap", route.ProfileHeapHandler())
	debug.GET("/pprof/goroutine", route.ProfileGoroutineHandler())
	debug.GET("/pprof/block", route.ProfileBlockHandler())
	debug.GET("/pprof/threadcreate", route.ProfileThreadCreateHandler())
	debug.GET("/pprof/cmdline", route.ProfileCmdlineHandler())
	debug.GET("/pprof/profile", route.ProfileProfileHandler())
	debug.GET("/pprof/symbol", route.ProfileSymbolHandler())
	debug.POST("/pprof/symbol", route.ProfileSymbolHandler())
	debug.GET("/pprof/trace", route.ProfileTraceHandler())

	path.GET("/favicon.ico", func(c *gin.Context) {
		c.Redirect(http.StatusMovedPermanently, "/static/favicon/images/favicon.ico")
	})

	path.GET("/manifest.json", func(c *gin.Context) {
		c.Redirect(http.StatusMovedPermanently, "/static/favicon/manifest.json")
	})

	path.GET("/browserconfig.xml", func(c *gin.Context) {
		c.Redirect(http.StatusMovedPermanently, "/static/favicon/browserconfig.xml")
	})

	path.GET("/", route.Index)
}

func initMiddleware(engine *gin.Engine, accessLogConfig *util.LogConfig) {
	engine.Use(logging.AccessLog(accessLogConfig))
	engine.Use(recovery.RecoveryJSON())

	// not found handler
	engine.NoRoute(recovery.RecoveryHttpError(http.StatusNotFound))
	engine.NoMethod(recovery.RecoveryHttpError(http.StatusMethodNotAllowed))

}
