package route

import (
	"os"
	"runtime"
	"time"

	"amuz.es/gogs/infra/api_skeleton/subsys/http/iface"
	"amuz.es/gogs/infra/api_skeleton/subsys/http/middleware/logging"
	"amuz.es/gogs/infra/api_skeleton/util"
	"github.com/gin-gonic/gin"
)

var (
	startTime          = time.Now()
)

func Metric(c *gin.Context) {
	_, readable := c.Request.URL.Query()["readable"]
	if readable {
		memory := serializeReadableStat()
		util.DumpJSON(c, 200, memory)
	} else {
		memory := serializeMemoryStat()
		util.DumpJSON(c, 200, memory)
	}
}

func serializeMemoryStat() *iface.Metric {
	m := new(runtime.MemStats)
	runtime.ReadMemStats(m)
	return &iface.Metric{
		Cmdline:      os.Args,
		Uptime:       uint64(time.Now().UnixNano() - startTime.UnixNano()),
		Request:iface.RequestMetric{
			Duration:iface.DurationMetric{
				Min:       uint64(logging.DurationAspect.Min),
				Max:       uint64(logging.DurationAspect.Max),
				Mean:      uint64(logging.DurationAspect.Mean),
				Stdev:     uint64(logging.DurationAspect.Stdev),
				P90:       uint64(logging.DurationAspect.P90),
				P95:       uint64(logging.DurationAspect.P95),
				P99:       uint64(logging.DurationAspect.P99),
				Timestamp: logging.DurationAspect.Timestamp,
			},
			Count:*logging.CounterAspect,
		},
		Runtime:iface.RuntimeMetric{
			GoVersion:    runtime.Version(),
			GoOs:         runtime.GOOS,
			GoArch:       runtime.GOARCH,
			CpuNum:       runtime.NumCPU(),
			GoroutineNum: runtime.NumGoroutine(),
			Gomaxprocs:   runtime.GOMAXPROCS(0),
			CgoCallNum:   runtime.NumCgoCall(),
		},
		Memory:iface.MemoryMetric{
			MemAllocated: m.Alloc,
			MemTotal:     m.TotalAlloc,
			MemSys:       m.Sys,
			Lookups:      m.Lookups,
			MemMallocs:   m.Mallocs,
			MemFrees:     m.Frees,

			HeapAlloc:    m.HeapAlloc,
			HeapSys:      m.HeapSys,
			HeapIdle:     m.HeapIdle,
			HeapInuse:    m.HeapInuse,
			HeapReleased: m.HeapReleased,
			HeapObjects:  m.HeapObjects,

			StackInuse:  m.StackInuse,
			StackSys:    m.StackSys,
			MSpanInuse:  m.MSpanInuse,
			MSpanSys:    m.MSpanSys,
			MCacheInuse: m.MCacheInuse,
			MCacheSys:   m.MCacheSys,
			BuckHashSys: m.BuckHashSys,
			GCSys:       m.GCSys,
			OtherSys:    m.OtherSys,

			NextGC:       m.NextGC,
			LastGC:       uint64(time.Now().UnixNano()) - m.LastGC,
			PauseTotalNs: m.PauseTotalNs,
			PauseNs:      m.PauseNs[(m.NumGC + 255) % 256],
			NumGC:        m.NumGC,
		},
	}
}

//
func serializeReadableStat() *iface.MetricReadable {
	m := new(runtime.MemStats)
	runtime.ReadMemStats(m)
	return &iface.MetricReadable{
		Uptime:       util.TimeSincePro(startTime),
		Request:iface.RequestMetricReadable{
			Duration:iface.DurationMetricReadable{
				Min:       util.TimeSinceDuration(uint64(logging.DurationAspect.Min)),
				Max:       util.TimeSinceDuration(uint64(logging.DurationAspect.Max)),
				Mean:      util.TimeSinceDuration(uint64(logging.DurationAspect.Mean)),
				Stdev:     util.TimeSinceDuration(uint64(logging.DurationAspect.Stdev)),
				P90:       util.TimeSinceDuration(uint64(logging.DurationAspect.P90)),
				P95:       util.TimeSinceDuration(uint64(logging.DurationAspect.P95)),
				P99:       util.TimeSinceDuration(uint64(logging.DurationAspect.P99)),
				Timestamp: logging.DurationAspect.Timestamp,
			},
			Count:*logging.CounterAspect,
		},
		Runtime:iface.RuntimeMetric{
			GoVersion:    runtime.Version(),
			GoOs:         runtime.GOOS,
			GoArch:       runtime.GOARCH,
			CpuNum:       runtime.NumCPU(),
			GoroutineNum: runtime.NumGoroutine(),
			Gomaxprocs:   runtime.GOMAXPROCS(0),
			CgoCallNum:   runtime.NumCgoCall(),
		},
		Memory:iface.MemoryMetricReadable{
			MemAllocated: util.FileSize(m.Alloc),
			MemTotal:     util.FileSize(m.TotalAlloc),
			MemSys:       util.FileSize(m.Sys),
			Lookups:      m.Lookups,
			MemMallocs:   m.Mallocs,
			MemFrees:     m.Frees,

			HeapAlloc:    util.FileSize(m.HeapAlloc),
			HeapSys:      util.FileSize(m.HeapSys),
			HeapIdle:     util.FileSize(m.HeapIdle),
			HeapInuse:    util.FileSize(m.HeapInuse),
			HeapReleased: util.FileSize(m.HeapReleased),
			HeapObjects:  m.HeapObjects,

			StackInuse:  util.FileSize(m.StackInuse),
			StackSys:    util.FileSize(m.StackSys),
			MSpanInuse:  util.FileSize(m.MSpanInuse),
			MSpanSys:    util.FileSize(m.MSpanSys),
			MCacheInuse: util.FileSize(m.MCacheInuse),
			MCacheSys:   util.FileSize(m.MCacheSys),
			BuckHashSys: util.FileSize(m.BuckHashSys),
			GCSys:       util.FileSize(m.GCSys),
			OtherSys:    util.FileSize(m.OtherSys),

			NextGC:       util.FileSize(m.NextGC),
			LastGC:       util.TimeSinceDuration(uint64(time.Now().UnixNano()) - m.LastGC) + " ago",
			PauseTotalNs: util.TimeSinceDuration(m.PauseTotalNs),
			PauseNs:      util.TimeSinceDuration(m.PauseNs[(m.NumGC + 255) % 256]),
			NumGC:        m.NumGC,
		},
	}
}
