package periodic

import (
	"time"

	cronLib "github.com/robfig/cron"

	"sync"

	"sync/atomic"

	"amuz.es/gogs/infra/api_skeleton/util"
	"context"
)

var (
	logger          = util.NewLogger("periodic")
	runningJobGroup = sync.WaitGroup{}
	runningJobCount = int32(0)
)

type Job struct {
	started   time.Time
	Name      string
	Execution func()
	isRunning int32
}

func (j *Job) Run() {
	if ok := atomic.CompareAndSwapInt32(&j.isRunning, 0, -1); !ok {
		return
	}

	// ENTER CRITICAL SECTION
	j.started = time.Now()
	runningJobGroup.Add(1)
	atomic.AddInt32(&runningJobCount, 1)
	defer func() {
		// EXIT CRITICAL SECTION
		atomic.AddInt32(&runningJobCount, -1)
		runningJobGroup.Done()
		atomic.StoreInt32(&j.isRunning, 0)
		logger.Debugf("Job %s Stopped in %s", j.Name, time.Now().Sub(j.started))
	}()
	j.Execution()
}

func Start(errorSignal chan error, ctx context.Context, waiter *sync.WaitGroup) {
	cron, err := initialize(waiter)
	if err != nil {
		errorSignal <- err
		return
	}
	go gracefulCloser(cron, ctx, waiter)
	logger.Info("scheduler started")
}

func initialize(waiter *sync.WaitGroup) (cron *cronLib.Cron, err error) {
	defer func() {
		if recov := recover(); recov != nil {
			err = recov.(error)
		}
	}()
	cron = cronLib.New()
	waiter.Add(1)
	logger.Debug("starting scheduler")
	defer cron.Start()
	return
}

func gracefulCloser(cron *cronLib.Cron, ctx context.Context, waiter *sync.WaitGroup) {
	<-ctx.Done()
	defer waiter.Done()
	logger.Info("stopping scheduler")
	cron.Stop()

	waitChannel := make(chan struct{}, 1)
	go func() {
		runningJobGroup.Wait()
		close(waitChannel)
	}()
	warningTicker := time.Tick(1 * time.Second)
	for {
		select {
		case <-waitChannel:
			logger.Info("Periodic Jobs shutdowned safely.")
			return
		case <-warningTicker:
			logger.Infof(
				"Waiting Periodic Jobs (remains %d job(s))",
				runningJobCount,
			)
		}
	}
}

func initJobs(closeSignal chan struct{}) {
	// cron.AddJob("* * * * * *", &Job{
	// 	Name: "job 1",
	// 	Execution: func() {
	// 		logger.Debugf("job 1 (executing %d job(s))", runningJobCount)
	// 		select {
	// 		case <-closeSignal:
	// 			logger.Debugf("job 1 quit signal received!")
	// 		case <-time.After(time.Millisecond * 2000):
	// 		}
	// 	},
	// })
	// todo upload 실패한건, 오래된건 관리.
}
