package redis

import (
	"errors"
	"fmt"
	"strconv"

	"amuz.es/gogs/infra/api_skeleton/util/iface"
	"github.com/garyburd/redigo/redis"
)

var (
	tokenStoreBaseKey        []byte
	refreshTokenStoreBaseKey []byte
	sessionStoreBaseKey      []byte
	tokenExpireTime          int64
	refreshTokenExpireTime   int64
)

func AppTokenInitService(phase string, tokenExpireTimeData, refreshTokenExpireTimeData int64) {
	tokenStoreBaseKey = []byte(fmt.Sprint(phase, "/RB", "/TK/"))
	refreshTokenStoreBaseKey = []byte(fmt.Sprint(phase, "/RB", "/RK/"))
	sessionStoreBaseKey = []byte(fmt.Sprint(phase, "/RB", "/SE/"))

	tokenExpireTime = tokenExpireTimeData / 1000
	refreshTokenExpireTime = refreshTokenExpireTimeData / 1000
}

func AppTokenCreateSession(clientId iface.UUID, memberId uint64) (out *SessionData, err error) {
	var client resourceConn
	if client, err = getConn(); err != nil {
		return
	}
	defer putConn(client)

	var (
		refreshToken, _ = iface.NewUUIDV4()
		token, _        = iface.NewUUIDV4()

		tokenStoreKey        = append(tokenStoreBaseKey, token.String()...)
		refreshTokenStoreKey = append(refreshTokenStoreBaseKey, refreshToken.String()...)
		sessionStoreKey      = strconv.AppendUint(sessionStoreBaseKey, memberId, 10)

		refreshTokenData []byte
		tokenData        []byte
		sessionData      []byte

		newSession = &SessionData{
			Token:        token,
			RefreshToken: refreshToken,
		}
	)

	if refreshTokenData, err = (&RefreshTokenData{
		ClientId: clientId,
		MemberId: memberId,
	}).Marshal(); err != nil {
		return
	}

	if tokenData, err = (&TokenData{
		ClientId:     clientId,
		MemberId:     memberId,
		RefreshToken: refreshToken,
	}).Marshal(); err != nil {
		return
	}
	if sessionData, err = newSession.Marshal(); err != nil {
		return
	}
	var prevSession = new(SessionData)

	var tx = func() (err error) {
		if _, err = client.Do("WATCH", sessionStoreKey); err != nil {
			return
		}

		if fetchErr := extractDataFromRedis(prevSession, sessionStoreKey, client); fetchErr != nil {
			prevSession = nil
		}
		client.Send("MULTI")

		if prevSession != nil {
			previousTokenStoreKey := append(tokenStoreBaseKey, prevSession.Token.String()...)
			previousRefreshTokenStoreKey := append(refreshTokenStoreBaseKey, prevSession.RefreshToken.String()...)
			client.Send("DEL", previousTokenStoreKey, previousRefreshTokenStoreKey)
		}
		client.Send("SETEX", tokenStoreKey, tokenExpireTime, tokenData)
		client.Send("SETEX", refreshTokenStoreKey, refreshTokenExpireTime, refreshTokenData)
		client.Send("SETEX", sessionStoreKey, refreshTokenExpireTime, sessionData)

		if _, err = redis.Values(client.Do("EXEC")); err != nil {
			return newTxError(err)
		}

		out = newSession
		return
	}

	err = retries(tx)
	return
}

func AppTokenFindMemberIdByToken(clientId iface.UUID, token iface.UUID) (memberId uint64, err error) {
	var client resourceConn
	if client, err = getConn(); err != nil {
		return
	}
	defer putConn(client)
	var (
		tokenStoreKey    = append(tokenStoreBaseKey, token.String()...)
		matchedTokenData = new(TokenData)
	)

	if err = extractDataFromRedis(matchedTokenData, tokenStoreKey, client); err != nil {
		return
	}

	if clientId.Equal(matchedTokenData.ClientId) {
		memberId = matchedTokenData.MemberId
	} else {
		client.Send("DEL", tokenStoreKey)
		err = errors.New("matched token is invalid")
	}
	return

}

func AppTokenFindMemberIdByRefreshToken(clientId iface.UUID, refreshToken iface.UUID) (memberId uint64, err error) {
	var client resourceConn
	if client, err = getConn(); err != nil {
		return
	}
	defer putConn(client)
	var (
		refreshTokenStoreKey    = append(refreshTokenStoreBaseKey, refreshToken.String()...)
		matchedRefreshTokenData = new(RefreshTokenData)
	)

	if err = extractDataFromRedis(matchedRefreshTokenData, refreshTokenStoreKey, client); err != nil {
		return
	}

	if clientId.Equal(matchedRefreshTokenData.ClientId) {
		memberId = matchedRefreshTokenData.MemberId
	} else {
		sessionStoreKey := strconv.AppendUint(sessionStoreBaseKey, matchedRefreshTokenData.MemberId, 10)
		sessionData := new(SessionData) //newSessionData
		if errFetchSession := extractDataFromRedis(sessionData, sessionStoreKey, client); errFetchSession != nil {
			logger.Info("failed to AppTokenFindMemberIdByRefreshToken", errFetchSession)
			client.Send("DEL", refreshTokenStoreKey, sessionStoreKey)
		} else {
			tokenStoreKey := append(tokenStoreBaseKey, sessionData.Token.String()...)
			client.Send("DEL", refreshTokenStoreKey, sessionStoreKey, tokenStoreKey)
		}
		err = errors.New("matched token is invalid")
	}
	return

}
func AppTokenRefreshSessionByRefreshToken(clientId iface.UUID, refreshToken iface.UUID) (out *SessionData, err error) {
	var client resourceConn
	if client, err = getConn(); err != nil {
		return
	}
	defer putConn(client)

	var (
		refreshTokenStoreKey = append(refreshTokenStoreBaseKey, refreshToken.String()...)
		prevRefreshTokenData = new(RefreshTokenData) //refreshTokenData
		prevSessionData      = new(SessionData)      //previousSessionData
		prevTokenData        = new(TokenData)        //previousTokenData

		memberId                        uint64
		sessionStoreKey                 []byte
		prevTokenStoreKey               []byte
		prevAnotherRefreshTokenStoreKey []byte
	)

	if fetchErr := extractDataFromRedis(prevRefreshTokenData, refreshTokenStoreKey, client); fetchErr != nil {
		err = errors.New("matched refreshToken not found")
		return
	} else {
		memberId = prevRefreshTokenData.MemberId
		sessionStoreKey = strconv.AppendUint(sessionStoreBaseKey, memberId, 10)
	}

	var tx = func() (err error) {
		if _, err = client.Do("WATCH", sessionStoreKey); err != nil {
			return
		}

		if err = extractDataFromRedis(prevSessionData, sessionStoreKey, client); err != nil {
			client.Send("MULTI")
			client.Send("DEL", refreshTokenStoreKey)
			if _, err = redis.Values(client.Do("EXEC")); err != nil {
				logger.Info("failed to AppTokenRefreshSessionByRefreshToken", err)
			}
			err = errors.New("matched sessionData not found")
			return
		} else {
			prevTokenStoreKey = append(tokenStoreBaseKey, prevSessionData.Token.String()...)
			prevAnotherRefreshTokenStoreKey = append(refreshTokenStoreBaseKey, prevSessionData.RefreshToken.String()...)
		}

		if fetchErr := extractDataFromRedis(prevTokenData, prevTokenStoreKey, client); fetchErr != nil {
			prevTokenData = nil
		}

		if prevRefreshTokenData.ClientId.Equal(clientId) &&
			prevSessionData.RefreshToken.Equal(refreshToken) &&
			(prevTokenData == nil || refreshToken.Equal(prevTokenData.RefreshToken)) {
			// 검증성공!
			var (
				newToken, _      = iface.NewUUIDV4()                               //token
				newTokenStoreKey = append(tokenStoreBaseKey, newToken.String()...) //tokenData

				newTokenData   []byte //tokenData
				newSessionData []byte //newSessionData
				newSession     = &SessionData{ //newSessionData
					Token:        newToken,
					RefreshToken: refreshToken,
				}
			)
			//newSession
			if newTokenData, err = (&TokenData{
				ClientId:     clientId,
				MemberId:     memberId,
				RefreshToken: refreshToken,
			}).Marshal(); err != nil {
				return
			}

			if newSessionData, err = newSession.Marshal(); err != nil {
				return
			}

			client.Send("MULTI")

			if prevTokenData != nil {
				client.Send("DEL", prevTokenStoreKey)
			}

			client.Send("SETEX", newTokenStoreKey, tokenExpireTime, newTokenData)
			client.Send("SETEX", sessionStoreKey, refreshTokenExpireTime, newSessionData)
			client.Send("EXPIRE", refreshTokenStoreKey, refreshTokenExpireTime)

			if _, err = redis.Values(client.Do("EXEC")); err != nil {
				return
			}

			out = newSession
			return
		} else {
			// 부정사용의혹!

			client.Send("MULTI")
			// 부정사용의혹!
			if prevTokenData != nil {
				client.Send("DEL", refreshTokenStoreKey, sessionStoreKey, prevTokenStoreKey, prevAnotherRefreshTokenStoreKey)
			} else {
				client.Send("DEL", refreshTokenStoreKey, sessionStoreKey, prevAnotherRefreshTokenStoreKey)
			}
			if _, err = redis.Values(client.Do("EXEC")); err != nil {
				logger.Info("failed to AppTokenRefreshSessionByRefreshToken", err)
			}

			err = errors.New("cannot refresh token")
		}
		return
	}

	err = retries(tx)
	return
}
