package iface

import (
	"bytes"
	"database/sql/driver"
	"encoding/hex"
	"errors"
	"github.com/google/uuid"
	"fmt"
)

func (u UUID) Marshal() ([]byte, error) {
	return u.UUID[:], nil
}

func (u UUID) MarshalTo(buf []byte) (n int, err error) {
	copy(buf, u.UUID[:])
	return len(u.UUID), nil
}

func (u *UUID) Unmarshal(buf []byte) error {
	if len(buf) != UUIDLength {
		return fmt.Errorf("invalid bytes16 (got %d bytes)", len(buf))
	}
	copy(u.UUID[:], buf)
	return nil
}

func (u *UUID) UnmarshalHex(buf []byte) (err error) {
	hexLen := hex.DecodedLen(len(buf))
	if hexLen != UUIDLength {
		return fmt.Errorf("invalid UUID (got %d bytes)", len(buf))

	}
	_, err = hex.Decode(u.UUID[:], buf)
	return
}

func (u *UUID) MarshalHex() ([]byte, error) {
	hexLen := hex.EncodedLen(UUIDLength)
	hexBuf := make([]byte,hexLen)
	if encodedLen := hex.Encode(hexBuf, u.UUID[:]); encodedLen != hexLen {
		return nil, fmt.Errorf("invalid bytes16 length: %d", len(hexBuf))
	}
	return hexBuf, nil
}

func (u UUID) MarshalJSON() (encoded []byte, err error) {
	encoded, err = u.MarshalHex()
	if err != nil {
		return
	}
	var buffer bytes.Buffer
	buffer.WriteRune('"')
	buffer.Write(encoded)
	buffer.WriteRune('"')
	return buffer.Bytes(), nil
}
func (u *UUID) UnmarshalJSON(from []byte) error {
	quote := []byte("\"")
	quoteSize := len(quote)

	if len(from) < quoteSize*2 {
		return errors.New("invalid quote notation")
	}

	if !bytes.HasPrefix(from, quote) || !bytes.HasSuffix(from, quote) {
		return errors.New("invalid quote notation")
	}
	return u.UnmarshalHex(from[quoteSize:len(from)-quoteSize])
}

func (u UUID) Compare(other UUID) int {
	return bytes.Compare(u.UUID[:], other.UUID[:])
}

func (u UUID) Equal(other UUID) bool {
	return u.Compare(other) == 0
}

func (u *UUID) Size() int {
	if u == nil {
		return 0
	}
	return UUIDLength
}

func NewUUIDV4() (u UUID, err error) {
	var ui uuid.UUID
	if ui, err = uuid.NewRandom(); err == nil {
		u.UUID = ui
	}
	return
}

// Scan implements the Scanner interface.
func (u *UUID) Scan(src interface{}) error {
	switch src.(type) {
	case string:
		u.UnmarshalHex([]byte(src.(string)))
	default:
		return errors.New("Incompatible type for UUID")
	}
	return nil
}

// Value implements the driver Valuer interface.
func (u UUID) Value() (driver.Value, error) {
	return u.MarshalHex()
}

// Value implements the driver Valuer interface.
func (u UUID) Random() {
	u.UUID = uuid.New()
}
